BodyController = {}
-- Nested object to make sure quick save will keep state
BodyController_info = {
    {
        -- list of all body controllers ingame
        bodyControllersList = {},
        -- Stores the index of the body controller with the best fitness
        bestFitnessIndex = nil,
        -- Stores the average fitness of all body controllers
        averageFitness = 0,
        -- Number of samples for the sliding fitness average
        averageMaxSamples = 100,
    }
}

function BodyController.run()
    local bcInfo = BodyController_info[1]
    --> Process instances.
    bcInfo.bestFitnessIndex = nil
    bcInfo.averageFitness = 0
    for i = 1, #bcInfo.bodyControllersList do
        local bodyController = bcInfo.bodyControllersList[i]
        bodyController.success = BodyController.brain.process(bodyController)
        bodyController.lastBodyCenter = AabbGetBodyCenterPos(bodyController.body)
        bcInfo.averageFitness = bcInfo.averageFitness + BodyController.fitness(bodyController)
    end
    if #bcInfo.bodyControllersList > 0 then
        bcInfo.averageFitness = bcInfo.averageFitness / #bcInfo.bodyControllersList
    end

    -- Run Genetic algorithm + Deep Q-Learning algorithm main loop
    genetic_algo.run()

    -- Apply highlights if needed
    if regGetBool('tool.highlightControllers') then

        for i = 1, #bcInfo.bodyControllersList do
            -- By default highlight is white
            local color = {r=1, g=1, b=1, a=1}
            if not bcInfo.bodyControllersList[i].success then
                -- Becomes red if controller had an error
                color.g = 0
                color.b = 0
            elseif not bcInfo.bodyControllersList[i].active then
                -- Becomes grey if body is inactive
                color.r = 0.5
                color.g = 0.5
                color.b = 0.5
            elseif i == bcInfo.bestFitnessIndex then
                -- Becomes green if body has the best fitness for current objective
                color.r = 0
                color.b = 0
            end
            -- Becomes half transparent if body objective is not selected objective
            local indexes = BodyController.getEligiblePopulationIndexes()
            if not TableFind(indexes, i) then
                color.a = 0.5
            end
            HighlightFullBody(bcInfo.bodyControllersList[i].body, color.r, color.g, color.b, color.a)
        end
    end

    --> Safely delete specified instances.
    BodyController.instances.removeMarkedBodyControllers()

end

function BodyController.resetFitness(bodyController)
    bodyController.fitnessSamples = {}
end

function BodyController.recordFitness(bodyController, fitness)
    local bcInfo = BodyController_info[1]
    TablePushFront(bodyController.fitnessSamples, fitness, bcInfo.averageMaxSamples)
end

function BodyController.getLastFitness(bodyController)
    if #bodyController.fitnessSamples == 0 then
        return 0
    else
        return bodyController.fitnessSamples[1]
    end
end

-- The fitness of a body controller is the average of each instant fitness recorded
function BodyController.fitness(bodyController)
    local fitnessCount = #bodyController.fitnessSamples
    if fitnessCount == 0 then
        return 0
    else
        local sum = 0
        for i=1,fitnessCount do
            sum = sum + bodyController.fitnessSamples[i]
        end
        return sum / fitnessCount
    end
end

-- Gives a new neural network to an existing body controller
function BodyController.setNeuralNetwork(bodyController, neuralNetwork)
    if not ai.neural_network.IsCompatible(neuralNetwork, bodyController.objective) then
        dbm("Cannot update neural network of body controller "..bodyController.id..": Incopatible objective")
        return
    end

    bodyController.neuralNetwork = neuralNetwork
    BodyController.resetFitness(bodyController)
    BodyController.activate(bodyController, regGetBool('tool.createActiveAI'))
end

-- Gives a new objective and neural network to an existing body controller
function BodyController.convert(bodyController, objective, neuralNetwork)
    local gaInfo = genetic_algo_info[1]
    bodyController.objective = objective
    BodyController.setNeuralNetwork(bodyController, neuralNetwork)
    BodyController.initRunnigAlgosData(bodyController)
end

-- Gets the fitness of the best body controller
function BodyController.bestFitness()
    local bcInfo = BodyController_info[1]
    if bcInfo.bestFitnessIndex == nil or bcInfo.bestFitnessIndex > #bcInfo.bodyControllersList then
        return 0
    else
        return BodyController.fitness(bcInfo.bodyControllersList[bcInfo.bestFitnessIndex])
    end
end

-- Gets the average fitness of all body controllers
function BodyController.averageFitness()
    local bcInfo = BodyController_info[1]
    return bcInfo.averageFitness
end

-- Gets the indexes of body controllers matching the current objective
function BodyController.getEligiblePopulationIndexes(objective)
    local bcInfo = BodyController_info[1]
    local indexes = {}
    objective = objective or regGetString('neural_network.objective')
    for i=1,#bcInfo.bodyControllersList do
        if bcInfo.bodyControllersList[i].objective == objective then
            TableAppend(indexes, i)
        end
    end
    return indexes
end

-- Randomize the neural network of this controller (used when training starts)
function BodyController.randomize(bodyController, deactivate)
    deactivate = deactivate or false
    ai.neural_network.Init(bodyController.neuralNetwork)
    BodyController.resetFitness(bodyController)
    BodyController.activate(bodyController, not deactivate)
end

-- Returns a flat list representing the genes of this controller
function BodyController.getGenes(bodyController)
    return ai.neural_network.GetFlatParams(bodyController.neuralNetwork)
end

-- Set the internal values of this controller based on the given gene list
function BodyController.setGenes(bodyController, genes)
    bodyController.success = ai.neural_network.SetFlatParams(bodyController.neuralNetwork, genes)
    BodyController.resetFitness(bodyController)
    BodyController.activate(bodyController, true)
end

-- Activate or deactivate a body controller
function BodyController.activate(bodyController, active)
    bodyController.active = active
end

-- Toggle active state of a body controller
function BodyController.toggleActivation(bodyController)
    bodyController.active = not bodyController.active
end

-- Int the parameters involved in Deep Q-Learning algorithm
function BodyController.initDeepQLearningParams(bodyController)
    bodyController.targetNeuralNetwork = DeepCopy(bodyController.neuralNetwork)
    bodyController.bufferNeuralNetwork = DeepCopy(bodyController.neuralNetwork)
    bodyController.trainingStep = 0
    BodyController.bufferClear(bodyController.buffer)
end

-- If genetic algo is running, randomize the newly created neural network
function BodyController.initRunnigAlgosData(bodyController)
    local gaInfo = genetic_algo_info[1]
     -- If genetic algo is running, reandomize the newly created neural network
     if gaInfo.state ~= "stopped" then
        -- Only randomize if the option is enabled
        if regGetBool('tool.randomAddDuringTrain') then
            BodyController.randomize(bodyController, gaInfo.state == "paused")
        else
            BodyController.activate(bodyController, gaInfo.state ~= "paused")
        end
        -- Initialize data specific to the Deep Q-Learning algo
        BodyController.initDeepQLearningParams(bodyController)
    end
end

-- Push a record at the end of the buffer
function BodyController.bufferPush(buffer, state_in, action, next_state, reward, done)
    TableAppend(buffer.states, DeepCopy(state_in))
    TableAppend(buffer.actions, DeepCopy(action))
    TableAppend(buffer.next_states, DeepCopy(next_state))
    TableAppend(buffer.rewards, DeepCopy(reward))
    TableAppend(buffer.dones, DeepCopy(done))
end

-- Clear the records buffer
function BodyController.bufferClear(buffer)
    buffer.states = {}
    buffer.actions = {}
    buffer.next_states = {}
    buffer.rewards = {}
    buffer.dones = {}
end

-- Return the curret size of the buffer
function BodyController.bufferCount(buffer)
    return #(buffer.states)
end






BodyController.instances = {}

--- Create a body controller object and store it in the list of instances.
function BodyController.instances.attach(body)
    local bcInfo = BodyController_info[1]

    if not IsBodyDynamic(body) then
        dbm("Cannot attach AI to object: The object is static")
        return
    end
    
    if IsBodyJointedToStatic(body) then
        dbm("Cannot attach AI to object: The object is connected to a static object")
        return
    end
    
    local bodyController = {
        id = #bcInfo.bodyControllersList + 1,
        removeBodyController = false,
        success = true,
        active = regGetBool('tool.createActiveAI'),
        lastBodyMass = GetBodyMass(body),
        objective = regGetString('neural_network.objective'),
        neuralNetwork = ai.neural_network.FromJson(regGetString('neural_network.current')),
        body = body,
        lastBodyCenter = AabbGetBodyCenterPos(body),
        fitnessSamples = {},
        -- Deep Q-Learning parameters
        lastSenseData = {{}, {}},
        lastAction = {0, 0},
        targetNeuralNetwork = nil,
        bufferNeuralNetwork = nil,
        buffer = {
            states = {},
            actions = {},
            next_states = {},
            rewards = {},
            dones = {},
        },
        trainingStep = 0,
    }

    if bodyController.neuralNetwork == nil then
        dbm("Cannot attach AI to object: Syntax error in neural network")
        return
    end

    if TableFind(objectives, bodyController.objective) == nil then
        dbm("Cannot attach AI to object: Unknown objective '"..bodyController.objective.."'")
        return
    end

    if not ai.neural_network.IsCompatible(bodyController.neuralNetwork, bodyController.objective) then
        dbm("Cannot attach AI to object: Incopatible objective")
        return
    end

    BodyController.initRunnigAlgosData(bodyController)
    TableAppend(bcInfo.bodyControllersList, bodyController)

end


--- Marks for removal the body controller associated with the specified body, return true if the controller existed and was removed
function BodyController.instances.detach(body)

    local bodyController = BodyController.instances.getBodyController(body)
    if bodyController then
        BodyController.instances.markBodyControllerForRemoval(bodyController)
        return true
    end
    return false

end

--- Attach to the body if it has no controller yet, else etach from it
function BodyController.instances.attachDetach(body)

    if not BodyController.instances.detach(body) then
        BodyController.instances.attach(body)
    end

end

--- Returns the body controller attached to the given body, or nil if not found
function BodyController.instances.getBodyController(body)

    local bcInfo = BodyController_info[1]
    for i=1,#bcInfo.bodyControllersList do
        if bcInfo.bodyControllersList[i].body == body then
            return bcInfo.bodyControllersList[i]
        end
    end
    return nil

end

--- Mark a body controller for removal to safely remove it near the end of the frame.
function BodyController.instances.markBodyControllerForRemoval(body_controller)
    body_controller.removeBodyController = true
end

--- Safely delete all body controllers that are marked for removal.
function BodyController.instances.removeMarkedBodyControllers()

    local bcInfo = BodyController_info[1]
    TableRemove(bcInfo.bodyControllersList, function(t, i, j)
        -- Refresh ids to match actual indexes
        if not bcInfo.bodyControllersList[i].removeBodyController then
            bcInfo.bodyControllersList[i].id = j
        end
        -- Return true to keep the value, or false to discard it.
        return not bcInfo.bodyControllersList[i].removeBodyController
    end)
    
end






BodyController.brain = {}
-- Map of actions to execute if the corresponding output of the neural network is the highest
function BodyController.brain.buildActionsMap()
    local map = {}
    -- Builds a map of every possible combination of [-1, 0, 1] for a 3D vector
    for i=-1,1 do
        for j=-1,1 do
            for k=-1,1 do
                TableAppend(map, Vec(i, j, k))
            end
        end
    end
    return map
end
BodyController.brain.actionsMap = BodyController.brain.buildActionsMap()

-- Read the inputs, feed them to the neural network, apply the output actions, update fitness
function BodyController.brain.process(bodyController)

    local alive = BodyController.brain.checkAlive(bodyController)
    if not alive then
        return false
    end
    local active = BodyController.brain.checkActivation(bodyController)
    if not active then
        -- Inactive AIs have 0 fitness
        BodyController.recordFitness(bodyController, 0)
        -- Keep the success value when inactive
        return bodyController.success
    end
    local inputs = BodyController.brain.sense(bodyController)
    if not inputs then
        return false
    end
    local outputs = ai.neural_network.ForwardPropagate(bodyController.neuralNetwork, inputs)
    if not outputs then
        return false
    end
    -- Select the action based on the epsilon-greedy policy (policy disabled when training is not running)
    local action = training.DeepQLearning.SelectEpsilonGreedyAction(outputs)
    local success = BodyController.brain.control(bodyController, action)
    if not success then
        return false
    end
    BodyController.brain.updateFitness(bodyController)
    return true
end

-- Check if the body is "alive" = if it's not fully broken, and if it is, remove its controller
function BodyController.brain.checkAlive(bodyController)

    if not IsBodyValid(bodyController.body) then
        dbm("Object controlled by AI "..bodyController.id.." was destroyed, removing AI")
        BodyController.instances.markBodyControllerForRemoval(bodyController)
        return false
    end
    return true

end

-- Check if the AI should remain active or become inactive
function BodyController.brain.checkActivation(bodyController)

    -- Inactive AIs remain inactive
    if not bodyController.active then
        return false
    end
    -- Chek if an error occured
    if regGetBool('neural_network.inactiveOnError') then
        if not bodyController.success then
            bodyController.active = false
            return false
        end 
    end
    -- Check if the object controlled was broken since last frame
    if regGetBool('neural_network.inactiveOnBreak') then
        local mass = GetBodyMass(bodyController.body)
        if mass < bodyController.lastBodyMass then
            bodyController.active = false
            bodyController.lastBodyMass = mass
            return false
        end 
    end
    return true

end

-- Read the list of inputs based on objective
function BodyController.brain.sense(bodyController)

    local senseData = {}
    local angularVel = GetBodyAngularVelocity(bodyController.body)
    if bodyController.objective == "balance" then
        -- 9 inputs: Body rotation on each axis (3) + Body rotation speed on each axis (3) + Body dimensions (3)
        local tr = GetBodyTransform(bodyController.body)
        local rx, ry, rz = GetQuatEuler(tr.rot)
        TableAppend(senseData, rx)
        TableAppend(senseData, ry)
        TableAppend(senseData, rz)
        TableAppend(senseData, angularVel[1])
        TableAppend(senseData, angularVel[2])
        TableAppend(senseData, angularVel[3])
        local localTr, min, max = GetShapeObb(GetMainShape(bodyController.body))
        local dim = AabbDimensions(min, max)
        TableAppend(senseData, dim[1])
        TableAppend(senseData, dim[2])
        TableAppend(senseData, dim[3])

    elseif bodyController.objective == "follow" then
        -- 7 inputs: Body rotation speed on each axis (3) + Unit vector pointing at player (3) + Distance to player (1)
        TableAppend(senseData, angularVel[1])
        TableAppend(senseData, angularVel[2])
        TableAppend(senseData, angularVel[3])
        local objToPlayer = VecSub(GetPlayerCenter(), AabbGetBodyCenterPos(bodyController.body))
        local playerDir = VecNormalize(objToPlayer)
        TableAppend(senseData, playerDir[1])
        TableAppend(senseData, playerDir[2])
        TableAppend(senseData, playerDir[3])
        local distToPlayer = VecLength(objToPlayer)
        TableAppend(senseData, distToPlayer)

    elseif bodyController.objective == "avoid" then
        -- 11 inputs: Body rotation speed on each axis (3) + Distance to object detected in all horizontal directions (8)
        TableAppend(senseData, angularVel[1])
        TableAppend(senseData, angularVel[2])
        TableAppend(senseData, angularVel[3])
        local distances = BodyController.brain.scanSurroundings(bodyController)
        for i=1,#distances do
            TableAppend(senseData, distances[i])
        end

    end
    TablePushFront(bodyController.lastSenseData, senseData)
    return senseData

end

-- Scan the surroundings and return a list of distances to objects hit
function BodyController.brain.scanSurroundings(bodyController)

    local distances = {}
    local direction = Vec(1, 0, 0)
    local center = AabbGetBodyCenterPos(bodyController.body)
    -- Raytrace in 8 directions
    for i=1,8 do
        local maxDist = regGetInt('neural_network.scanRange')
        QueryRejectBody(bodyController.body)
        local jointedBodies = GetJointedBodies(bodyController.body)
        for j=1,#jointedBodies do
            QueryRejectBody(jointedBodies[j])
        end
        local hit, dist, normal, shape = QueryRaycast(center, direction, maxDist)
        if not hit then
            dist = maxDist
        end
        TableAppend(distances, dist)
        if db then
            local hitPoint = VecAdd(center, VecResize(direction, dist))
            local maxPoint = VecAdd(center, VecResize(direction, maxDist))
            dbl(center, hitPoint, 0, 0, 1, 1)
            dbl(hitPoint, maxPoint, 1, 0, 0, 1)
        end
        -- Rotate aim by 45 degs
        direction = VecRotate(direction, "y", 45)
    end
    return distances

end

-- Control the body with the given action values
function BodyController.brain.control(bodyController, action)
    -- Gets the action vector from the action index
    local actionVec = BodyController.brain.actionsMap[action]
    -- Action vector is a vector with each component in range [-1, 0, 1], indicating if we want to reduce, keep or accelerate speed on the given angle
    local currentAngularVel = SetBodyAngularVelocity(bodyController.body)
    local newAngularVel = VecAdd(currentAngularVel, VecResize(actionVec, regGetFloat('neural_network.accelRate') * 100.0 * GetTimeStep()))
    SetBodyAngularVelocity(bodyController.body, newAngularVel)
    -- Record last action for DeepQLearning algo
    TablePushFront(bodyController.lastAction, action)
    return true
end

-- Compute the fitness for the current time step and record its value
-- The fitness is stored as a percentage of how good is the object doing to reach the objective
function BodyController.brain.updateFitness(bodyController)

    local bcInfo = BodyController_info[1]
    local fitness = 0
    if bodyController.objective == "balance" then
        -- When balancing:
        -- - the worst possible position is when the smallest side is flat down
        -- - the best is when the diagonal between opposite angles is straight up
        local tr, min, max = GetShapeObb(GetMainShape(bodyController.body))
        local dim = AabbDimensions(min, max)
        local corners = GetObbCorners(tr, min, max)
        local lowBound = math.min(dim[1], dim[2], dim[3])
        local highBound = VecLength(dim)
        local lowY = nil
        local highY = nil
        for i=1,#corners do
            local y = corners[i][2]
            if lowY == nil or y < lowY then
                lowY = y
            end
            if highY == nil or y > highY then
                highY = y
            end
        end
        local height = highY - lowY
        fitness = ratio(height, lowBound, highBound)
        if db then
            ObbDraw(corners, 0, 0, 1, 1)
            if bodyController.body == tool.tool.aimBody then
                ui.debugStr = "dim="..VecStr(dim)..", lowBound="..lowBound..", highBound="..highBound..", lowY="..lowY..", highY="..highY..", height="..height
            end
        end
        
    elseif bodyController.objective == "follow" then
        -- When following (far from player):
        -- - the worst is when it moved away from the player since last step
        -- - the best is when it moved in the direction of the player since last step
        -- When following (close to player):
        -- - the worst is when it has a high rotation speed
        -- - the best is when it has a low rotation speed
        local bodyCenter = AabbGetBodyCenterPos(bodyController.body)
        local playerCenter = GetPlayerCenter()
        local distToPlayer = VecDist(bodyCenter, playerCenter)
        local minDistToPlayer = regGetInt('neural_network.distToPlayer')
        if distToPlayer > minDistToPlayer then
            local movementDir = VecSub(bodyCenter, bodyController.lastBodyCenter)
            local playerDir = VecSub(playerCenter, bodyCenter)
            -- Dot is 1 if facing same direction, 0 if orthogonal, -1 if facing opposite direction
            local dot = VecDot(VecNormalize(movementDir), VecNormalize(playerDir))
            -- Makes sure not moving is interpreted as moving orthogonally
            local speed = VecLength(movementDir)
            if speed < 0.0001 then
                dot = 0
            end
            -- Bring this [-1, 1] range back to [0, 1]
            fitness = ratio(dot, -1, 1)
            if db then
                local dir = VecResize(movementDir, 1)
                dbl(bodyCenter, VecAdd(bodyCenter, dir), 0, 0, 1, 1)
                dbl(playerCenter, bodyCenter, 0, 1, 0, 1)
                if bodyController.body == tool.tool.aimBody then
                    ui.debugStr = "distToPlayer="..distToPlayer..", minDistToPlayer="..minDistToPlayer..", speed="..speed..", dot="..dot
                end
            end
        else
            local angularSpeed = VecLength(GetBodyAngularVelocity(bodyController.body))
            -- Brings value between 0 and 1, with 1 when speed is close to 0 and 0 when speed goes to infinite in any direction
            fitness = 2 / (math.exp(angularSpeed) + math.exp(-angularSpeed))
            if db then
                if bodyController.body == tool.tool.aimBody then
                    ui.debugStr = "distToPlayer="..distToPlayer..", minDistToPlayer="..minDistToPlayer..", angularSpeed="..angularSpeed
                end
            end
        end
       
       
    elseif bodyController.objective == "avoid" then
        -- When avoiding:
        -- - the worst is when an item is very close to the object in any direction
        -- - the best is when no item is near the object in any direction
        local distances = BodyController.brain.scanSurroundings(bodyController)
        local minDist = nil
        for i=1,#distances do
            local dist = distances[i]
            if minDist == nil or dist < minDist then
                minDist = dist
            end
        end
        fitness = ratio(minDist, 0, regGetInt('neural_network.scanRange'))
        if db then
            if bodyController.body == tool.tool.aimBody then
                ui.debugStr = "minDist="..minDist
            end
        end

    end

    BodyController.recordFitness(bodyController, fitness)
    -- Store best fitness for this cycle and the current objective
    if bodyController.objective == regGetString('neural_network.objective') and  BodyController.fitness(bodyController) > BodyController.bestFitness() then
        bcInfo.bestFitnessIndex = bodyController.id
    end

end

-- Returns the result of the last action with the associated reward (fitness)
function BodyController.brain.getLastStep(bodyController)
    -- State of previous turn produced action of previous turn
    local state_in = ternary(#bodyController.lastSenseData[2] > 0, bodyController.lastSenseData[2], BodyController.brain.getDummyInput(bodyController))
    local action = bodyController.lastAction[2] or 0
    -- Current state and reward is the result of this action
    local next_state = ternary(#bodyController.lastSenseData[1] > 0, bodyController.lastSenseData[1], BodyController.brain.getDummyInput(bodyController))
    local reward = BodyController.getLastFitness(bodyController) * 100
    local validData = bodyController.active and (action ~= 0)
    local done = not validData
    -- print("====================")
    -- print("state_in="..json.encode(state_in))
    -- print("action="..json.encode(action))
    -- print("next_state="..json.encode(next_state))
    -- print("reward="..json.encode(reward))
    -- print("validData="..json.encode(validData))
    -- print("done="..json.encode(done))
    return state_in, action, next_state, reward, done
end

-- Returns an empty set of inputs
function BodyController.brain.getDummyInput(bodyController)
    local inputs = {}
    local actualInputs, actualOutputs = ai.neural_network.GetInOut(bodyController.neuralNetwork)
    for i=1,actualInputs do
        TableAppend(inputs, 0)
    end
    return inputs
end