

genetic_algo = {}
-- Nested object to make sure quick save will keep state
genetic_algo_info = {
    {
        -- stopped/running/paused
        state = "stopped",
        generation = 1,
        timer = 0,
        objective = "",
    }
}

function genetic_algo.start()
    local bcInfo = BodyController_info[1]
    local gaInfo = genetic_algo_info[1]
    gaInfo.objective = regGetString('neural_network.objective')
    -- Check if there is enough population to start
    local indexesToTrain = BodyController.getEligiblePopulationIndexes(gaInfo.objective)
    if #indexesToTrain < 1 then
        dbm("Needs at least one AI with current objective to start training")
        return
    end

    for i=1,#indexesToTrain do
        local bodyController = bcInfo.bodyControllersList[indexesToTrain[i]]
        -- Randomize weights and bias of the whole population to train
        BodyController.randomize(bodyController)
        -- Make bodies unbreakable if needed
        if regGetBool('genetic_algo.unbreakableTrain') then
            SetTag(bodyController.body, "unbreakable")
        end
        -- Init the Depp Q-Learning algorithm controller level variables 
        BodyController.initDeepQLearningParams(bodyController)
    end

    -- Init the Depp Q-Learning and Adam algorithm global level variables 
    training.DeepQLearning.Reset()
    ai.optim.Adam.Reset()
    
    gaInfo.timer = regGetInt('genetic_algo.generationDuration')
    gaInfo.generation = 1
    gaInfo.state = "running"
end

function genetic_algo.stop()
    local bcInfo = BodyController_info[1]
    local gaInfo = genetic_algo_info[1]
    local indexesToTrain = BodyController.getEligiblePopulationIndexes(gaInfo.objective)
    for i=1,#indexesToTrain do
        local bodyController = bcInfo.bodyControllersList[indexesToTrain[i]]
        -- Make bodies breakable again if needed
        if regGetBool('genetic_algo.unbreakableTrain') then
            RemoveTag(bodyController.body, "unbreakable")
        end
    end

    -- Ends the Depp Q-Learning algorithm
    training.DeepQLearning.Terminate()

    gaInfo.state = "stopped"
end

function genetic_algo.pause()
    local bcInfo = BodyController_info[1]
    local gaInfo = genetic_algo_info[1]
    local indexesToTrain = BodyController.getEligiblePopulationIndexes(gaInfo.objective)
    for i=1,#indexesToTrain do
        local bodyController = bcInfo.bodyControllersList[indexesToTrain[i]]
        -- Deactivate all AIs during pause
        BodyController.activate(bodyController, false)
    end

    gaInfo.state = "paused"
end


function genetic_algo.resume()
    local bcInfo = BodyController_info[1]
    local gaInfo = genetic_algo_info[1]
    local indexesToTrain = BodyController.getEligiblePopulationIndexes(gaInfo.objective)
    for i=1,#indexesToTrain do
        local bodyController = bcInfo.bodyControllersList[indexesToTrain[i]]
        -- Reactivate all AIs on resume
        BodyController.activate(bodyController, true)
        -- Also reset fitness
        BodyController.resetFitness(bodyController)
    end

    -- Reset timer after resume
    gaInfo.timer = regGetInt('genetic_algo.generationDuration')
    gaInfo.state = "running"
end

function genetic_algo.toggleStart()
    local gaInfo = genetic_algo_info[1]
    if gaInfo.state == "stopped" then
        genetic_algo.start()
    else
        genetic_algo.stop()
    end
end

function genetic_algo.togglePause()
    local gaInfo = genetic_algo_info[1]
    if gaInfo.state == "running" then
        genetic_algo.pause()
    elseif gaInfo.state == "paused" then
        genetic_algo.resume()
    end
end

function genetic_algo.getState()
    local gaInfo = genetic_algo_info[1]
    return gaInfo.state
end

function genetic_algo.getGeneration()
    local gaInfo = genetic_algo_info[1]
    return gaInfo.generation
end

function genetic_algo.run()
    local bcInfo = BodyController_info[1]
    local gaInfo = genetic_algo_info[1]
    if gaInfo.state ~= "running" then
        return
    end

    -- Check if there is enough population to keep training
    local indexesToTrain = BodyController.getEligiblePopulationIndexes(gaInfo.objective)
    if #indexesToTrain < 1 then
        dbm("Training stopped: No more population")
        genetic_algo.stop()
        return
    end

    local genomes = {}
    for i=1,#indexesToTrain do
        local bodyController = bcInfo.bodyControllersList[indexesToTrain[i]]
        if regGetBool('genetic_algo.trainDuringGen') then
            -- Update the Deep Q-Learning algorithm
            local state_in, action, next_state, reward, done = BodyController.brain.getLastStep(bodyController)
            local stepProgressed = training.DeepQLearning.Update(bodyController.neuralNetwork, bodyController.targetNeuralNetwork, bodyController.bufferNeuralNetwork, bodyController.buffer, state_in, action, next_state, reward, done)
            -- if an error occured
            if stepProgressed == nil then
                dbm("Training paused: Error occured during training")
                genetic_algo.pause()
                return
            end
            if stepProgressed then
                bodyController.trainingStep = bodyController.trainingStep + 1
            end
        end
    end

    -- Updates the variables that needs to move at the end of each loop
    training.DeepQLearning.UpdateGlobal()

    -- Let the population evolve freely for the intended duration
    if gaInfo.timer > 0 then
        gaInfo.timer = gaInfo.timer - GetTimeStep()
        return
    end
    
    -- Build a list of genomes sorted by fitness (best fitness first)
    local genomes = {}
    for i=1,#indexesToTrain do
        local bodyController = bcInfo.bodyControllersList[indexesToTrain[i]]
        local genFit = {
            fitness = BodyController.fitness(bodyController),
            genome = BodyController.getGenes(bodyController)
        }
        TableAppend(genomes, genFit)
    end
    table.sort(genomes, function(a,b) return a.fitness > b.fitness end)

    -- Main steps of the genetic algorithm: selection / crossover / mutation
    local newGenomes = genetic_algo.selection(genomes)
    --print("newGenomes1="..json.encode(newGenomes))
    -- newGenomes is used as in-out param
    genetic_algo.crossover(newGenomes)
    --print("newGenomes2="..json.encode(newGenomes))
    -- newGenomes is used as in-out param
    genetic_algo.mutation(newGenomes)
    --print("newGenomes3="..json.encode(newGenomes))
    -- Give the new genomes generated to a random body controllers
    for i=1,#newGenomes do
        local randIndex = rdm(1, #indexesToTrain)
        local bodyController = bcInfo.bodyControllersList[indexesToTrain[randIndex]]
        --print("nbGenes "..i.."="..#newGenomes[i].genome)
        BodyController.setGenes(bodyController, newGenomes[i].genome)
        -- When training time is over, update the target NN from the main NN trained by the Deep Q-Learning algorithm
        bodyController.targetNeuralNetwork = DeepCopy(bodyController.neuralNetwork)
        table.remove(indexesToTrain, randIndex)
    end

    -- Go to next generation
    gaInfo.generation = gaInfo.generation + 1
    gaInfo.timer = regGetInt('genetic_algo.generationDuration')
end

-- Select the population members that "survived" for the next generation
function genetic_algo.selection(genomes)
    -- Keep only the top X% of the population to pick parents for the next generation
    -- Also compute total fitness for the remaining genomes
    local genomeIndex = #genomes
    local totalPopulation = #genomes
    local totalFitness = 0
    local keepTopPercent = regGetInt('genetic_algo.keepTopPercent')
    while genomeIndex > 0 do
        -- When top percentage is reached, stop removing values from the list and start counting total fitness
        if (100.0 * #genomes / totalPopulation) <= keepTopPercent then
            totalFitness = totalFitness + genomes[genomeIndex].fitness
        -- Else remove values until keepTopPercent is reached
        else
            table.remove(genomes, genomeIndex)
        end
        genomeIndex = genomeIndex - 1
    end

    -- Compute cumulated fitness frequency (index in this list are the same as in genomes)
    local cumulFitnessFreqs = {}
    local cumul = 0
    for i=1,#genomes do
        cumul = cumul + (genomes[i].fitness / totalFitness)
        TableAppend(cumulFitnessFreqs, cumul)
    end

    -- Pick the genomes of the new population according to the given selection type
    local newGenomes = {}
    local selectionType = regGetString('genetic_algo.selectionType')
    local nbAgentsToSelect = totalPopulation
    if selectionType == "linear" then
        -- Make a linear repartition of all agents over the cumulative fitness list
        local index = 1
        local freqCursor = 0.0
        local freqStep = 1.0 / nbAgentsToSelect
        for i=1,nbAgentsToSelect do
            while index < #cumulFitnessFreqs and cumulFitnessFreqs[index] < freqCursor  do
                index = index + 1
            end
            TableAppend(newGenomes, genomes[index])
            freqCursor = freqCursor + freqStep
        end
    elseif selectionType == "roulette" then
        -- Take random picks in the cumulative fitness list, respectng the proportion of each parent genome
        for i=1,nbAgentsToSelect do
            local fitIndex = TableBinarySearch(cumulFitnessFreqs, rdmf())
            TableAppend(newGenomes, genomes[fitIndex])
        end
    end
    return newGenomes
end

-- Make crossover between some memebers of the population and produce childs
function genetic_algo.crossover(newGenomes)
    local crossoverGenomes = {}
    local newGenomesCount = #newGenomes;
    local crossoverPercent = regGetInt('genetic_algo.crossoverPercent')
    local crossoverGenomesCount = floatToInt(newGenomesCount * crossoverPercent / 100.0)
    -- Select genomes for crossover and remove them from the list of new genomes
    while #crossoverGenomes < crossoverGenomesCount do
        local genomeIdx = rdm(0, #newGenomes)
        TableAppend(crossoverGenomes, newGenomes[genomeIdx])
        table.remove(newGenomes, genomeIdx)
    end
    -- Cross random genomes from the pool until the initial population is reached
    local crossoverType = regGetString('genetic_algo.crossoverType')
    while #newGenomes < newGenomesCount do
        local parent1Index = rdm(1, #crossoverGenomes)
        local parent2Index = rdm(1, #crossoverGenomes)
        local parent1Genes = crossoverGenomes[parent1Index].genome
        local parent2Genes = crossoverGenomes[parent2Index].genome
        local childGenes = {}
        local slicePosition1 = rdm(1, #parent1Genes)
        local slicePosition2 = rdm(1, #parent1Genes)
        local minSlicePosition = math.min(slicePosition1, slicePosition2)
        local maxSlicePosition = math.max(slicePosition1, slicePosition2)
        -- For each gene, pick the one from the correct parent based on the method selected
        for i=1,#parent1Genes do
            if crossoverType ==  "slice-1" then
                if i <= slicePosition1 then
                    TableAppend(childGenes, parent1Genes[i])
                else
                    TableAppend(childGenes, parent2Genes[i])
                end
            elseif crossoverType ==  "slice-2" then
                if i <= minSlicePosition then
                    TableAppend(childGenes, parent1Genes[i])
                elseif i <= maxSlicePosition then
                    TableAppend(childGenes, parent2Genes[i])
                else
                    TableAppend(childGenes, parent1Genes[i])
                end
            elseif crossoverType ==  "random" then
                if rdm() then -- 50% chance to be 0 or 1
                    TableAppend(childGenes, parent1Genes[i])
                else
                    TableAppend(childGenes, parent2Genes[i])
                end
            end
        end
        -- Add the child back to the list of new genomes
        local childGenome = {
            fitness = 0,
            genome = childGenes,
        }
        TableAppend(newGenomes, childGenome)
    end
    return newGenomes
end

-- Mutate one gene for some members of the population
function genetic_algo.mutation(newGenomes)
    local mutationRate = regGetInt('genetic_algo.mutationRate')
    local mutationMin = regGetInt('genetic_algo.mutationMin')
    local mutationMax = regGetInt('genetic_algo.mutationMax')
    local mutationMultiply = regGetBool('genetic_algo.mutationMultiply')
    for i=1,#newGenomes do
        -- If this agent should mutate
        if rdmf(0, 100) <= mutationRate then
            -- Randomly mutate one gene
            local mutationIndex = rdm(1, #newGenomes[i].genome)
            local mutationValue = rdmf(mutationMin, mutationMax)
            if mutationMultiply then
                newGenomes[i].genome[mutationIndex] = newGenomes[i].genome[mutationIndex] * mutationValue
            else
                newGenomes[i].genome[mutationIndex] = mutationValue
            end
        end
    end
    return newGenomes
end

-- Update the unbreakable state of bodies based on setting and current training state
function genetic_algo.updateUnbreakableStates()
    local bcInfo = BodyController_info[1]
    local gaInfo = genetic_algo_info[1]
    local indexesToTrain = BodyController.getEligiblePopulationIndexes(gaInfo.objective)
    for i=1,#indexesToTrain do
        local bodyController = bcInfo.bodyControllersList[indexesToTrain[i]]
        -- Make bodies breakable/unbreakable if needed
        if regGetBool('genetic_algo.unbreakableTrain') and gaInfo.state == "running" then
            SetTag(bodyController.body, "unbreakable")
        else
            RemoveTag(bodyController.body, "unbreakable")
        end
    end
end