-- Script inspired by the following examples, adapted and completed with features needed for the AI Trainer mod:
-- https://github.com/Dragonrick123/Neural-Network-in-lua-Module/blob/general/MainModule.lua (lots of errors in this one but I kept the main code structure)
-- https://gist.github.com/cassiozen/de0dff87eb7ed599b5d0 (probably a bit more functional, but harder to read)
-- NB: This might not be the most optimized way to store and process a neural network, so don't expect crazy performance out of it.
--     The primary goal here whas to get something easy enough to read so I can understand a little bit what I'm doing :p

ai = {
    neural_network = {}
}

-- Neural network structure example (as generated through the ai.neural_network.AddLayer() function)
ai.neural_network.example = {
    -- Number of inputs of the network (Neurons in the first layer)
    nbInputs = 3,
    -- Array of Layers
    layers = {
        -- Layer 0 (input) is not stored as it has no weight or bias
        -- Layer 1
        {
            -- Activation function of this layer, must match with an existing function name in ai.activations
            activation = "relu",
            -- Array of Neurons
            neurons = {
                -- Neuron 1
                {
                    -- One weight per Neuron in Layer 0
                    weights = {0, 0, 0},
                    -- One bias for the current Neuron
                    bias = 0,
                },
                -- Neuron 2
                {
                    weights = {0, 0, 0},
                    bias = 0,
                }
            }
        },
        -- ...
        -- As many layers as needed can be added here, with a custom number of neurons and activation function each
        -- ...
        -- Layer 2 (output)
        {
            activation = "sigmoid",
            neurons = {
                -- Neuron 1
                {
                    -- One weight per neuron in Layer 1
                    weights = {0, 0},
                    bias = 0,
                },
                -- Neuron 2
                {
                    weights = {0, 0},
                    bias = 0,
                },
                -- Neuron 3
                {
                    weights = {0, 0},
                    bias = 0
                },
                -- Neuron 4
                {
                    weights = {0, 0},
                    bias = 0,
                }
            }
        }
    }
}


ai.optim = {
    Adam = {}
}
-- Nested object to make sure quick save will keep state
ai_optim_Adam = {
    {
        m_dparams = {},
        v_dparams = {},
        beta1 = 0.9,
        beta2 = 0.999,
        epsilon = 1e-8,
        eta = 0.01,
    }
}

-- Resets the Adam optimization algorithm
function ai.optim.Adam.Reset()
    local self = ai_optim_Adam[1]
	self.m_dparams = {}
    self.v_dparams = {}
    self.t_dparams = {}
    self.beta1 = 0.9
    self.beta2 = 0.999
    self.epsilon = 1e-8
    self.eta =  0.01
end

-- Runs one step of the Adam optimization algorithm for the given parameter value
-- Use paramName to keep the previous information of each parameter to optimize
function ai.optim.Adam.Update(paramName, value, gradient)
	local self = ai_optim_Adam[1]
    if self.m_dparams[paramName] == nil then
        self.m_dparams[paramName] = 0
    end
    if self.v_dparams[paramName] == nil then
        self.v_dparams[paramName] = 0
    end
    if self.t_dparams[paramName] == nil then
        self.t_dparams[paramName] = 1
    end
    self.m_dparams[paramName] = self.beta1*self.m_dparams[paramName] + (1-self.beta1)*gradient
    self.v_dparams[paramName] = self.beta2*self.v_dparams[paramName] + (1-self.beta2)*(gradient^2)
    local m_dparam_corr = self.m_dparams[paramName]/(1-self.beta1^self.t_dparams[paramName])
    local v_dparam_corr = self.v_dparams[paramName]/(1-self.beta2^self.t_dparams[paramName])
    self.t_dparams[paramName] = self.t_dparams[paramName] + 1
    return value - self.eta*(m_dparam_corr/(math.sqrt(v_dparam_corr)+self.epsilon))
end


ai.activations = {
    derivate = {}
}
-- I found some useful hints on what type of activation function to use for what type of problem here:
-- https://towardsdatascience.com/how-to-choose-the-right-activation-function-for-neural-networks-3941ff0e6f9c
-- limits: [-inf, +inf], curve: /
function ai.activations.identity(array)
	return array
end
function ai.activations.derivate.identity(array)
    local out = {}
    for i=1,#array do
        out[i] = 1
    end
	return out
end

-- limits: [-1, 1], curve: _/‾
function ai.activations.tanh(array)
    local out = {}
    for i=1,#array do
        out[i] = math.tanh(array[i])
    end
	return out
end
function ai.activations.derivate.tanh(array)
    local out = {}
    for i=1,#array do
        out[i] = 1 - (math.tanh(array[i])^2)
    end
	return out
end

-- limits: [0, 1], curve: _/‾
function ai.activations.sigmoid(array)
    local out = {}
    for i=1,#array do
        out[i] = 1/(1 + math.exp(-array[i]))
    end
	return out
end
function ai.activations.derivate.sigmoid(array)
    local sigmoids = ai.activations.sigmoid(array)
    local out = {}
    for i=1,#array do
        out[i] = sigmoids[i] * (1-sigmoids[i])
    end
	return out
end

-- limits: [0, +inf], curve: _/
function ai.activations.relu(array)
    local out = {}
    for i=1,#array do
        out[i] = math.max(0, array[i])
    end
	return out
end
function ai.activations.derivate.relu(array)
    local out = {}
    for i=1,#array do
        out[i] = ternary(array[i] > 0, 1, 0)
    end
	return out
end

-- limits: [-inf, +inf], curve: _/
function ai.activations.leaky_relu(array)
    local out = {}
    local alpha = 0.01
    for i=1,#array do
        out[i] = math.max(alpha*array[i], array[i])
    end
	return out
end
function ai.activations.derivate.leaky_relu(array)
    local out = {}
    for i=1,#array do
        -- Normally derivate is undefined at point 0, but keep it at 0.01 anyway
        out[i] = ternary(array[i] > 0, 1, 0.01)
    end
	return out
end

-- limits: [0, 6], curve: _/‾
function ai.activations.relu6(array)
    local out = {}
    for i=1,#array do
        out[i] = math.min(math.max(0, array[i]), 6)
    end
	return out
end
function ai.activations.derivate.relu6(array)
    local out = {}
    for i=1,#array do
        out[i] = ternary(array[i] > 0, ternary(array[i] < 6, 1, 0), 0)
    end
	return out
end

-- limits: [0, 1], curve: /\/\/
function ai.activations.softmax(array)
    local sum = 0
    local exps = {}
    for i=1,#array do
        exps[i] = math.exp(array[i])
        sum = sum + exps[i]
    end
    local out = {}
    for i=1,#array do
        out[i] = exps[i]/sum
    end
	return out
end
function ai.activations.derivate.softmax(array)
    local out = {}
    for i=1,#array do
        out[i] = 1
    end
	return out
end

-- limits: [0, +inf], curve: _/
function ai.activations.swish(array)
	local sigmoids = ai.activations.sigmoid(array)
    local out = {}
    for i=1,#array do
        out[i] = sigmoids[i] * array[i]
    end
	return out
end
function ai.activations.derivate.swish(array)
	local sigmoids = ai.activations.sigmoid(array)
    local out = {}
    for i=1,#array do
        out[i] = math.exp(-array[i]) * (sigmoids[i]^2)
    end
	return out
end

-- limits: [0, +inf], curve: _/
function ai.activations.hard_swish(array)
    local out = {}
    for i=1,#array do
        out[i] = array[i] * (math.min(math.max(0, array[i] + 3), 6) / 6 )
    end
	return out
end
function ai.activations.derivate.hard_swish(array)
	local sigmoids = ai.activations.sigmoid(array)
    local out = {}
    for i=1,#array do
        out[i] = ternary(array[i] > 3, 1, ternary(array[i] > -3, (2*array[i] + 3)/6, 0))
    end
	return out
end

-- Returns the activation function to use based on its name
function ai.activations.GetFunction(Activation)
	if type(Activation) == "function" then
	    return Activation
    elseif ai.activations[Activation] then
		return ai.activations[Activation]
    else
        return nil
	end
end

-- Returns the derivate of the activation function to use based on its name
function ai.activations.GetDerivateFunction(Activation)
    if ai.activations.derivate[Activation] then
        return ai.activations.derivate[Activation]
    else
        return nil
    end
end

-- Creates an empty Neural Network with the given number of inputs
function ai.neural_network.Create(nbInputs)
	local newNeuralNetwork = {
        nbInputs = nbInputs,
        layers = {}
    }
    return newNeuralNetwork
end

-- Copy an existing Neural Network
function ai.neural_network.Copy(neuralNetwork)
	return DeepCopy(neuralNetwork)
end

-- Export a neural network to a json string
function ai.neural_network.ToJson(neuralNetwork)
    local status, retval = pcall(json.encode, neuralNetwork);
    if not status then
        dbm("Failed to export neural network to json: "..retval)
        return nil
    end
	return retval
end

-- Import a neural network from a json string
function ai.neural_network.FromJson(jsonStr)
    local status, retval = pcall(json.decode, jsonStr);
    if not status then
        dbm("Failed to import neural network from json: "..retval)
        return nil
    end
	return retval
end

-- Adds a layer at the end of an existing Neural Network
function ai.neural_network.AddLayer(neuralNetwork, nbNeurons, activation)
	local newLayer = {
        activation = activation,
        neurons = {}
    }
    local newNeuron = {
        weights = {},
        bias = 0,
    }
    -- Creates the weights array with the correct number of values bases on the number of neurons in the previous layer
    local neuronsInPreviousLayer = neuralNetwork.nbInputs
    local lastLayer = #neuralNetwork.layers
    if lastLayer > 0 then
        neuronsInPreviousLayer = #neuralNetwork.layers[lastLayer].neurons
    end
    for i=1,neuronsInPreviousLayer do
        TableAppend(newNeuron.weights, 0)
    end
    for i=1,nbNeurons do
        TableAppend(newLayer.neurons,  DeepCopy(newNeuron))
    end
    TableAppend(neuralNetwork.layers, newLayer)
end



------------------------------------------------------------------------------------------------------
-- START MOD-SPECIFIC (functions in this block are specific to this mod, outside of it are generic) --
------------------------------------------------------------------------------------------------------

-- Returns the default neural network to use for each training objective
function ai.neural_network.GetDefaultNN(objective)
    if objective == "balance" then
        -- 9 inputs: Body rotation on each axis (3) + Body rotation speed on each axis (3) + Body dimensions (3)
        local newNN = ai.neural_network.Create(9)
        -- 1 hidden layer of 9 neurons (arbitrary)
        ai.neural_network.AddLayer(newNN, 9, "leaky_relu")
        -- 27 outputs: All possible combination of action choices from BodyController.brain.actionsMap
        ai.neural_network.AddLayer(newNN, 27, "identity")
        return newNN
    elseif objective == "follow" then
        -- 7 inputs: Body rotation speed on each axis (3) + Unit vector pointing at player (3) + Distance to player (1)
        local newNN = ai.neural_network.Create(7)
        -- 1 hidden layers of 7 neurons (arbitrary)
        ai.neural_network.AddLayer(newNN, 7, "leaky_relu")
        -- 27 outputs: All possible combination of action choices from BodyController.brain.actionsMap
        ai.neural_network.AddLayer(newNN, 27, "identity")
        return newNN
    elseif objective == "avoid" then
        -- 11 inputs: Body rotation speed on each axis (3) + Distance to object detected in all horizontal directions (8)
        local newNN = ai.neural_network.Create(11)
        -- 1 hidden layers of 11 neurons (arbitrary)
        ai.neural_network.AddLayer(newNN, 11, "leaky_relu")
        -- 27 outputs: All possible combination of action choices from BodyController.brain.actionsMap
        ai.neural_network.AddLayer(newNN, 27, "identity")
        return newNN
    else
        dbm("GetDefaultNN - Unknown objective: "..tostring(objective))
        return nil
    end
end

-- Returns if the given neural network if compatible with the objective
function ai.neural_network.IsCompatible(neuralNetwork, objective)
    local expectedInputs, expectedOutputs = ai.neural_network.GetExpectedInOut(objective)
    local actualInputs, actualOutputs = ai.neural_network.GetInOut(neuralNetwork)
    return actualInputs == expectedInputs and actualOutputs == expectedOutputs
end

-- Returns the expected number of inputs and outputs for this objective
function ai.neural_network.GetExpectedInOut(objective)
    local expectedInputs = -1
    local expectedOutputs = -1
    if objective == "balance" then
        expectedInputs = 9
        expectedOutputs = 27
    elseif objective == "follow" then
        expectedInputs = 7
        expectedOutputs = 27
    elseif objective == "avoid" then
        expectedInputs = 11
        expectedOutputs = 27
    else
        dbm("GetExpectedInOut - Unknown objective: "..tostring(objective))
    end
    return expectedInputs, expectedOutputs
end

-- Returns the number of inputs and outputs for this neural network
function ai.neural_network.GetInOut(neuralNetwork)
    if not neuralNetwork then
        return nil, nil
    end
    local inputs = neuralNetwork.nbInputs
    local outputs = 0
    if #neuralNetwork.layers > 0 then
        outputs = #neuralNetwork.layers[#neuralNetwork.layers].neurons
    end
    return inputs, outputs
end

------------------------------------------------------------------------------------------------------
-- END MOD-SPECIFIC                                                                                 --
------------------------------------------------------------------------------------------------------



-- Computes the value returned by the given neuron of the given layer for the provided inputs
function ai.neural_network.NeuronThink(neuralNetwork, layerIndex, neuronIndex, inputs)
    if layerIndex < 1 or layerIndex > #neuralNetwork.layers then
        dbm("NeuronThink: Layer index out of bounds: "..tostring(layerIndex))
        return nil
    end
    local layer = neuralNetwork.layers[layerIndex]
    if neuronIndex < 1 or neuronIndex > #layer.neurons then
        dbm("NeuronThink: Neuron index out of bounds in layer "..tostring(layerIndex)..": "..tostring(neuronIndex))
        return nil
    end
    local neuron = layer.neurons[neuronIndex]
    if #inputs ~= #neuron.weights then
        dbm("NeuronThink: Mismatch between inputs and weights available for neuron "..neuronIndex.." of layer "..layerIndex..": "..tostring(#inputs).."/"..tostring(#neuron.weights))
        return nil
    end
    -- The value produced by a neuron is the sum of each input multiplied by its weight, plus the bias
    local res = TableMultiply(inputs, neuron.weights)
    local output = neuron.bias
    for i, v in pairs(res) do
        output = output + res[i]
    end
    return output
end

-- Apply the activation function of the layer (restrain the input to a predefined range)
function ai.neural_network.ActivateLayer(neuralNetwork, layerIndex, inputs)
    if layerIndex < 1 or layerIndex > #neuralNetwork.layers then
        dbm("ActivateLayer: Layer index out of bounds: "..tostring(layerIndex))
        return nil
    end
    local layer = neuralNetwork.layers[layerIndex]
    local activationFn = ai.activations.GetFunction(layer.activation)
    if activationFn == nil then
        dbm("ActivateLayer: '"..tostring(layer.activation).."' is not a supported activation function")
        return nil
    end
    return activationFn(inputs)
end

-- Propagates the given inputs through the whole network and returns the output values
function ai.neural_network.ForwardPropagate(neuralNetwork, inputs, returnAll)
    if #inputs ~= neuralNetwork.nbInputs then
        dbm("ForwardPropagate: Mismatching sizes nbInputs="..neuralNetwork.nbInputs..", #inputs="..#inputs)
        return nil
    end
    
    local allOutputs = {}
    local prevLayerOutputs = inputs
    for layerIndex=1,#neuralNetwork.layers do
        local layer = neuralNetwork.layers[layerIndex]
        local outputs = {}
        for neuronIndex=1,#layer.neurons do
            -- Compute the valeur of each neuron for the given input
            outputs[neuronIndex] = ai.neural_network.NeuronThink(neuralNetwork, layerIndex, neuronIndex, prevLayerOutputs)
            -- Abort in case of error
            if outputs[neuronIndex] == nil then
                return nil
            end
        end
        -- Apply the activation function to the above results
        outputs = ai.neural_network.ActivateLayer(neuralNetwork, layerIndex, outputs)
        -- Abort in case of error
        if outputs == nil then
            return nil
        end
        prevLayerOutputs = outputs
        TableAppend(allOutputs, outputs)
    end
    if returnAll then
        return allOutputs
    else
        return prevLayerOutputs
    end
end

-- Propagates the outputs through the whole network and update weights and bias to get closer to the provided inputs
function ai.neural_network.BackwardPropagate(neuralNetwork, inputs, desiredOutputs)

    local actualInputs, actualOutputs = ai.neural_network.GetInOut(neuralNetwork)
    if #inputs ~= actualInputs then
        dbm("BackwardPropagate: Mismatching sizes actualInputs="..actualInputs..", #inputs="..#inputs)
        return false
    end
    if #desiredOutputs ~= actualOutputs then
        dbm("BackwardPropagate: Mismatching sizes actualOutputs="..actualOutputs..", #desiredOutputs="..#desiredOutputs)
        return false
    end
    
    -- get the internal outputs for each layer
    local activations = ai.neural_network.ForwardPropagate(neuralNetwork, inputs, true)
    -- Get the gradients
    local param_grads = ai.neural_network.backward_step(neuralNetwork, inputs, activations, desiredOutputs)
    if param_grads == nil then
        return false
    end
    -- Update all flat params with the Adam algorithm
    local currentParams = ai.neural_network.GetFlatParams(neuralNetwork)
    local paramNames = ai.neural_network.GetUniqueParamNames(neuralNetwork)
    if #currentParams ~= #param_grads or #paramNames ~= #param_grads then
        dbm("BackwardPropagate: Mismatching sizes #currentParams="..#currentParams..", #paramNames="..#paramNames..", #param_grads="..#param_grads)
        return false
    end
    local updatedParams = {}
    for i=1,#currentParams do
        local param = currentParams[i]
        local newParam = ai.optim.Adam.Update(paramNames[i], currentParams[i], param_grads[i])
        TableAppend(updatedParams, newParam)
    end
    local success = ai.neural_network.SetFlatParams(neuralNetwork, updatedParams)
    if not success then
        return false
    end
    return true

end

-- Define the backward propagation step as a method
-- Inspired from https://blog.zhaytam.com/2018/08/15/implement-neural-network-backpropagation/
function ai.neural_network.backward_step(neuralNetwork, inputs, activations, targets)
    local param_grads = {}  -- List of parameter gradients for each layer
    local output_grad = {} -- The error gradient at the output of the current layer
    -- Initialize gradient form the network expected and actual outputs
    local networkOutputs = activations[#activations]
    for i=1,#targets do
        TableAppend(output_grad, networkOutputs[i] - targets[i])   
    end
    -- Propagate the error backwards through all the layers.
    local layers = neuralNetwork.layers
    -- Iterate backwards over the list of layers.
    for layerIndex=#layers,1,-1 do
        local layerOutputs = activations[layerIndex] -- Get the activations of the last layer on the stack
        -- Compute the error at the output layer.
        -- The output layer error is calculated different than hidden layer error.
        local input_grad = nil
        input_grad = ai.neural_network.get_input_grad(neuralNetwork, layerIndex, layerOutputs, output_grad)
        if input_grad == nil then
            dbm("backward_step: Error while processing layer "..layerIndex)
            return nil
        end
        -- Get the input of this layer (activations of the previous layer, or input values)
        local layerInputs = nil
        if layerIndex > 1 then
            layerInputs = activations[layerIndex-1]
        else
            layerInputs = inputs
        end
        -- Compute the layer parameter gradients used to update the parameters
        local grads = ai.neural_network.get_params_grad(neuralNetwork, layerInputs, output_grad)
        if grads == nil then
            dbm("backward_step: Error while processing layer "..layerIndex)
            return nil
        end
        param_grads = TableConcat(grads, param_grads)
        -- Compute gradient at output of previous layer (input of current layer):
        output_grad = input_grad
        
    end
    return param_grads  -- Return the parameter gradients
end

-- Return a list of gradients over the parameters
-- This builds a flat list list in the same order taken by ai.neural_network.SetFlatParams()
function ai.neural_network.get_params_grad(neuralNetwork, inputs, output_grad)
    local param_grads = {}
    for outputIndex=1,#output_grad do
        local bias_grad = 0
        for inputIndex=1,#inputs do
            local weight_grad = inputs[inputIndex] * output_grad[outputIndex]
            TableAppend(param_grads, weight_grad)
            bias_grad = bias_grad + output_grad[outputIndex]
        end
        TableAppend(param_grads, bias_grad)
    end
    return param_grads
end

-- Return the gradient at the inputs of this layer
function ai.neural_network.get_input_grad(neuralNetwork, layerIndex, outputs, output_grad)
    local input_grads = {}
    local layer = neuralNetwork.layers[layerIndex]
    local activationDerivateFn = ai.activations.GetDerivateFunction(layer.activation)
    if activationDerivateFn == nil then
        dbm("get_input_grad: No derivate fuction found for '"..tostring(layer.activation).."'")
        return nil
    end
    local derivates = activationDerivateFn(outputs)
    if #output_grad ~= #derivates then
        dbm("get_input_grad: Mismatching sizes #output_grad="..#output_grad..", #derivates="..#derivates)
        return nil
    end
    for neuronIndex=1,#outputs do
        local neuron = layer.neurons[neuronIndex]
        -- First compute gradient for the current activation function
        local activ_grad = output_grad[neuronIndex] * derivates[neuronIndex]
        for inputIndex=1,#neuron.weights do
            if input_grads[inputIndex] == nil then
                input_grads[inputIndex] = 0
            end
            -- Then the gradient for each weight
            local input_grad = activ_grad * neuron.weights[inputIndex]
            input_grads[inputIndex] = input_grads[inputIndex] + input_grad
        end
    end
    return input_grads
end

-- Randomly initializes a neural network using the "Xavier and He" initialisation method
-- https://prateekvishnu.medium.com/xavier-and-he-normal-he-et-al-initialization-8e3d7a087528
function ai.neural_network.Init(neuralNetwork)
	for layerIndex=1,#neuralNetwork.layers do
        local layer = neuralNetwork.layers[layerIndex]
        for neuronIndex=1,#layer.neurons do
            local neuron = layer.neurons[neuronIndex]
            neuron.bias = 0
            local nbWeights = #neuron.weights
            for weightIndex=1,nbWeights do
                neuron.weights[weightIndex] = rdmsd()*math.sqrt(2/nbWeights)
            end
        end
    end
end

-- Returns a flat list of all weights of the neural network
function ai.neural_network.GetFlatWeights(neuralNetwork)
    local weights = {}
    for layerIndex=1,#neuralNetwork.layers do
        local layer = neuralNetwork.layers[layerIndex]
        for neuronIndex=1,#layer.neurons do
            local neuron = layer.neurons[neuronIndex]
            for weightIndex=1,#neuron.weights do
                weights[#weights+1] = neuron.weights[weightIndex]
            end
        end
    end
    return weights
end

-- Returns a flat list of all biases of the neural network
function ai.neural_network.GetFlatBiases(neuralNetwork)
    local biases = {}
    for layerIndex=1,#neuralNetwork.layers do
        local layer = neuralNetwork.layers[layerIndex]
        for neuronIndex=1,#layer.neurons do
            biases[#biases+1] = layer.neurons[neuronIndex].bias
        end
    end
    return biases
end

-- Returns a flat list of all parameters of the neural network
function ai.neural_network.GetFlatParams(neuralNetwork)
    local params = {}
    -- This loop must match exactly with the one in ai.neural_network.SetFlatParams(neuralNetwork)
    for layerIndex=1,#neuralNetwork.layers do
        local layer = neuralNetwork.layers[layerIndex]
        for neuronIndex=1,#layer.neurons do
            local neuron = layer.neurons[neuronIndex]
            for weightIndex=1,#neuron.weights do
                TableAppend(params, neuron.weights[weightIndex])
            end
            TableAppend(params, neuron.bias)
        end
    end
    return params
end

-- Returns a flat list of uniaue param names
function ai.neural_network.GetUniqueParamNames(neuralNetwork)
    local paramNames = {}
    -- This loop must match exactly with the one in ai.neural_network.SetFlatParams(neuralNetwork)
    for layerIndex=1,#neuralNetwork.layers do
        local layer = neuralNetwork.layers[layerIndex]
        for neuronIndex=1,#layer.neurons do
            local neuron = layer.neurons[neuronIndex]
            for weightIndex=1,#neuron.weights do
                TableAppend(paramNames, tostring(neuralNetwork).."_L"..layerIndex.."_N"..neuronIndex.."_W"..weightIndex)
            end
            TableAppend(paramNames, tostring(neuralNetwork).."_L"..layerIndex.."_N"..neuronIndex.."_B")
        end
    end
    return paramNames
end

-- Loads a flat list of parameters in this neural network
function ai.neural_network.SetFlatParams(neuralNetwork, params)
    local tmpParams = ai.neural_network.GetFlatParams(neuralNetwork)
    if #tmpParams ~= #params then
        dbm("Cannot update neural network from flat parameters, expected "..#tmpParams.." parameters but got "..#params)
        return false
    end

    local paramIndex = 1
    local function getNextParam()
        local param = params[paramIndex]
        paramIndex = paramIndex + 1
        return param
    end
    -- This loop must match exactly with the one in ai.neural_network.GetFlatParams(neuralNetwork)
    for layerIndex=1,#neuralNetwork.layers do
        local layer = neuralNetwork.layers[layerIndex]
        for neuronIndex=1,#layer.neurons do
            local neuron = layer.neurons[neuronIndex]
            for weightIndex=1,#neuron.weights do
                neuron.weights[weightIndex] = getNextParam()
            end
            neuron.bias = getNextParam()
        end
    end
    return true
end