training = {
    DeepQLearning = {}
}
-- Nested object to make sure quick save will keep state
training_DeepQLearning = {
    {
        epsilon = 0.0,
        discount = 0.99,
    }
}

-- Perform a training iteration on a batch of data sampled from the experience replay buffer.
-- Algo bsed on this post: https://stackoverflow.com/questions/39932611/how-to-implement-deep-q-learning-gradient-descent
function training.DeepQLearning.TrainStep(main_nn, target_nn, states, actions, rewards, next_states, dones)
    local self = training_DeepQLearning[1]
    local buffer_size = #states
    for i=1,buffer_size do
        -- print("WWWWWWWWWWWWWWWWWWWWWW")
        -- print("states[i]="..json.encode(states[i]))
        -- print("actions[i]="..json.encode(actions[i]))
        -- print("rewards[i]="..json.encode(rewards[i]))
        -- print("next_states[i]="..json.encode(next_states[i]))
        -- print("dones[i]="..json.encode(dones[i]))

        -- If done flag is true, then we skip this sample
        if not dones[i] then
            -- Do a feedforward pass for the current state s to get predicted Q-values for all actions.
            local q_value = ai.neural_network.ForwardPropagate(main_nn, states[i])
            if not q_value then
                return false
            end
            -- Do a feedforward pass for the next state s’ to get predicted Q-values of the target network for all actions.
            local next_q_value = ai.neural_network.ForwardPropagate(target_nn, next_states[i])
            -- Number of possible actions is number of outputs of the network
            local num_actions = #q_value
            -- Calculate maximum overall network outputs max a’ Q(s’, a’).
            local max_next_q_value = 0
            for j=1,num_actions do
                local output = next_q_value[j]
                if output > max_next_q_value then
                    max_next_q_value = output
                end
            end
            -- Calculate target outputs
            local targets = {}
            for j=1,num_actions do
                -- Set the expected value for the action selected by the network
                if j == actions[i] then
                    -- Set Q-value target for action to r + γmax a’ Q(s’, a’) (use the max calculated above).
                    TableAppend(targets, rewards[i] + (self.discount * max_next_q_value))
                else
                    -- For all other actions, set the Q-value target to the same as originally returned from step 1, making the error 0 for those outputs
                    TableAppend(targets, q_value[j])
                end
            end
            -- Update the weights using backpropagation.
            local success = ai.neural_network.BackwardPropagate(main_nn, states[i], targets)
            if not success then
                return false
            end
        end
    end
    return true
end

-- Resets the Deep Q-Learning optimization algorithm
function training.DeepQLearning.Reset()
    local self = training_DeepQLearning[1]
	self.epsilon = 1.0
	self.discount = 0.99
end

-- Runs one step of the Deep Q-Learning optimization algorithm
function training.DeepQLearning.Update(main_nn, target_nn, buffer_nn, buffer, state_in, action, next_state, reward, done)
    local batch_size = regGetInt('training.batch_size')
    local batchMode = regGetBool('training.batchMode')
    -- Save to experience replay buffer.
    BodyController.bufferPush(buffer, state_in, action, next_state, reward, done)
    -- This is the indended way of working but it creates huge lag spikes when many AIs are trained at once
    if batchMode then
        -- When requested batch size is reached, train the neural network with the data recorded
        if BodyController.bufferCount(buffer) >= batch_size then
            -- Shuffle the samples to train a more generic network
            training.DeepQLearning.ShuffleSamples(buffer)
            -- Do one step of trainig using the collected samples
            local success = training.DeepQLearning.TrainStep(main_nn, target_nn, buffer.states, buffer.actions, buffer.rewards, buffer.next_states, buffer.dones)
            if not success then
                return nil
            end
            -- Remove all samples
            BodyController.bufferClear(buffer)
            return true
        end
    else
        -- Process the samples on the fly and store the results in a third neural network
        local success = training.DeepQLearning.TrainStep(buffer_nn, target_nn, {state_in}, {action}, {reward}, {next_state}, {done})
        if not success then
            return nil
        end
        -- When requested batch size is reached, update the main neural network
        if BodyController.bufferCount(buffer) >= batch_size then
            local trainedParams = ai.neural_network.GetFlatParams(buffer_nn)
            ai.neural_network.SetFlatParams(main_nn, trainedParams)
            -- Remove all samples
            BodyController.bufferClear(buffer)
            return true
        end
    end
    return false
end

-- Runs one global step of the Deep Q-Learning optimization algorithm (once after each body controller have been processed)
function training.DeepQLearning.UpdateGlobal()
    local self = training_DeepQLearning[1]
    -- Reduce epsilon value over time
    if self.epsilon > 0.05 then
        local decreaseRate = (10^(-regGetInt('training.epsilonDecrease')/4.0))/100.0
        self.epsilon = self.epsilon - decreaseRate -- default 0.001
    end
end

-- Shuffle the position of each element in the buffer (keeping coherence across all buffers)
function training.DeepQLearning.ShuffleSamples(buffer)
    local size = BodyController.bufferCount(buffer)
    for i=1,size do
        local j = rdm(i,size)
        TableSwapIndex(buffer.states, i, j)
        TableSwapIndex(buffer.actions, i, j)
        TableSwapIndex(buffer.rewards, i, j)
        TableSwapIndex(buffer.next_states, i, j)
        TableSwapIndex(buffer.dones, i, j)
    end
end

-- Take random action with probability epsilon, else take best action
function training.DeepQLearning.SelectEpsilonGreedyAction(outputs)
    local self = training_DeepQLearning[1]
    -- Roll random value in [0, 1]
    if rdmf() < training.DeepQLearning.GetExplorationRatio() then
        return rdm(1, #outputs) -- Random action.
    else
        return TableIndexOfMax(outputs) -- Greedy action for state.
    end
end

-- Completes the Deep Q-Learning optimization algorithm
function training.DeepQLearning.Terminate()
	local self = training_DeepQLearning[1]
	self.epsilon = 0.0
end

-- Returns the current exploration vs execution ratio
-- Exploration is when the AI makes decisions randomly, execution is whne it follows the best choice from the neural network
function training.DeepQLearning.GetExplorationRatio()
    local self = training_DeepQLearning[1]
    if regGetBool('genetic_algo.trainDuringGen') then
        return self.epsilon
    else
        return 0
    end
end